# This code is part of qtealeaves.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Two-body interactions in one-dimensional systems.
"""
import numpy as np

from .baseterm import _ModelTerm1D
from .sparsematrixoperator import SparseMatrixOperator


__all__ = ["TwoBodyTerm1D"]


class TwoBodyTerm1D(_ModelTerm1D):
    """
    The term defines an interaction between two sites of the Hilbert space.
    For example, the tunneling term in the Bose-Hubbard model can be
    represented by this term. This class represents the 1d version.

    **Arguments**

    operators : list of two strings
        String identifier for the operators. Before launching the simulation,
        the python API will check that the operator is defined.

    shift : int
        Defines the distance of the interaction in compliance
        with the systems in higher dimensions. In the end,
        we iterate over all sites and apply interactions to
        sites (x,) and (x + shift,)

    strength : str, callable, numeric (optional)
        Defines the coupling strength of the local terms. It can
        be parameterized via a value in the dictionary for the
        simulation or a function.
        Default to 1.

    prefactor : numeric, scalar (optional)
        Scalar prefactor, e.g., in order to take into account signs etc.
        Default to 1.

    add_complex_conjg : bool, optional
        (BUG ticket #1) Aims to automatically add complex conjugated
        terms, e.g., for the tunneling term.

    has_obc : bool or list of bools, optional
        Defines the boundary condition along each spatial dimension.
        If scalar is given, the boundary condition along each
        spatial dimension is assumed to be equal.
        Default to True
    """

    def __init__(
        self,
        operators,
        shift,
        strength=1,
        prefactor=1,
        add_complex_conjg=False,
        has_obc=True,
    ):
        self.operators = operators
        self.shift = shift
        self.strength = strength
        self.prefactor = prefactor
        self.add_complex_conjg = add_complex_conjg

        if isinstance(has_obc, bool):
            self.has_obc = [has_obc]
        else:
            self.has_obc = has_obc

    def iter_shifts(self):
        """
        Return all possible shifts, which is in the case of the
        1d systems exactly the defined shift.
        """
        yield self.shift

    def get_entries(self, params):
        """
        Entries based on two coordinates; one of them is the
        origin (0,).

        **Arguments**

        params : dictionary
            Contains the simulation parameters.

        **Details**

        To-Do: complex-conjugate terms are wrong, i.e., they work
        for sigma_{i}^{+} sigma_{j}^{-} or b_{i}^{dagger} b_{j}, but
        not for terms as b_{j} sigma_{i}^{x}!
        """
        for shift in self.iter_shifts():
            coord_a = 0
            coord_b = shift
            coordinates = [coord_a, coord_b]

            coupl_ii = {"coordinates": coordinates, "operators": self.operators}

            yield coupl_ii

            if self.add_complex_conjg:

                coupl_ii = {
                    "coordinates": coordinates,
                    "operators": self.operators[::-1],
                }

                yield coupl_ii

    def get_interactions(self, ll, params, **kwargs):
        """
        Iterate over all interaction. The iterator yields
        a dictionary with the operator keys and the generic
        shift based on the origin, and a list with the
        coordinates in the 1d system.
        """
        for elem in self.get_entries(params):
            for ix in self.iterate_sites(ll):
                coord_a, coord_b = elem["coordinates"]

                if np.sum(np.abs(np.array(coord_a))) != 0:
                    raise Exception("Coordinate A is not the origin.")

                jx = ix + coord_b

                if (jx >= ll[0]) and self.has_obc[0]:
                    continue

                if (jx < 0) and self.has_obc[0]:
                    continue

                if jx >= ll[0]:
                    jx = jx % ll[0]
                if jx < 0:
                    jx += ll[0]

                if (jx >= ll[0]) or (jx < 0):
                    raise Exception("Improve handling.")

                if ix == jx:
                    raise Exception("Same site ...")

                # Still in python indices
                coords_1d = [ix, jx]

                yield elem, coords_1d

    def get_fortran_str(self, ll, params, operator_map, param_map):
        """
        See :func:`_ModelTerm.get_fortran_str_two_body`.
        """
        return self.get_fortran_str_twobody(ll, params, operator_map, param_map)

    def get_sparse_matrix_operators(
        self, ll, params, operator_map, param_map, **kwargs
    ):
        """
        Construct the sparse matrix operator for this term.

        **Arguments**

        ll : int
            System size.

        params : dictionary
            Contains the simulation parameters.

        operator_map : OrderedDict
            For operator string as dictionary keys, it returns the
            corresponding integer IDs.

        param_map : dict
            This dictionary contains the mapping from the parameters
            to integer identifiers.

        kwargs : keyword arguments
            Keyword arguments are passed to `get_interactions`
        """
        if self.shift != 1:
            raise NotImplementedError(
                "Only nearest neighbor implemented for sparse MPO."
            )

        sp_mat_ops = []
        for ii in range(np.prod(ll)):
            sp_mat_ops.append(
                SparseMatrixOperator(ii == 0, ii + 1 == np.prod(ll), True)
            )

        op_id_0 = operator_map[(self.operators[0], "l")]
        op_id_1 = operator_map[(self.operators[1], "r")]

        param_repr = self.get_param_repr(param_map)

        for _, inds in self.get_interactions(ll, params, **kwargs):
            if abs(inds[0] - inds[1]) == np.prod(ll) - 1:
                raise Exception(
                    "Periodic boundary conditions not implemented for spMat."
                )

            if inds[0] == 0:
                shape_l = (1, 3)
            else:
                shape_l = (2, 3)

            if inds[1] + 1 == np.product(ll):
                shape_r = (3, 1)
            else:
                shape_r = (3, 2)

            sp_mat_l = np.zeros(shape_l, dtype=int)
            sp_mat_r = np.zeros(shape_r, dtype=int)

            prefactor_l = np.zeros(shape_l, dtype=np.complex128)
            prefactor_r = np.zeros(shape_r, dtype=np.complex128)

            params_l = np.zeros(shape_l, dtype=int)
            params_r = np.zeros(shape_r, dtype=int)

            sp_mat_l[-1, 1] = op_id_0
            prefactor_l[-1, 1] = self.prefactor
            params_l[-1, 1] = param_repr

            sp_mat_r[1, 0] = op_id_1
            prefactor_r[1, 0] = 1.0
            params_r[1, 0] = -1

            sp_mat_ops[inds[0]].add_term(sp_mat_l, params_l, prefactor_l)
            sp_mat_ops[inds[1]].add_term(sp_mat_r, params_r, prefactor_r)

        return sp_mat_ops
